/**
 * @file Vector data source implementation.
 * @license
 * Copyright (c) 2020 Felix Kopp <felix.kopp@stud.th-deg.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import { IVectorDataSource } from '../interfaces/vector-data-source';
import { IVector } from '../interfaces/vect';
import { fetchJSON } from '../util/xhr';

/** The amount of elements in the buffer below which new data is fetched. */
const BUF_THRESHOLD: number = 60;
/** How many data to fetch. */
const BUF_FETCH_SIZE: number = BUF_THRESHOLD * 2;

/**
 * Data source class for delivering a buffered stream of vector field data.
 */
export class VectorDataSource implements IVectorDataSource {

    /** The request uri to get the data from. */
    private uri: string;

    /**
     * The internal buffer for smooth and lag-free display (hopefully).
     * New elements are appended to the end, and data is taken from the start.
     */
    private buffer: IVector[][] = [];

    /** Whether we are refilling the cache right now. */
    private fetching: boolean = false;

    /**
     * Create a new data source.
     *
     * @param uri The request uri to get data from.
     */
    public constructor(uri: string) {
        this.uri = uri;
        this.fetch();
    }

    /** Refill the cache. */
    private async fetch() {
        for (let i: number = 0; i < BUF_FETCH_SIZE; i++) {
            try {
                const data: any = await fetchJSON(this.uri);
                this.buffer.push(data.arrows);
            } catch (err) {
                /** TODO: display an error message or something */
                console.error(err);
            }
        }
        this.fetching = false;
    }

    /**
     * @override
     * @inheritdoc
     */
    public getNext(): IVector[] | null {
        if (this.buffer.length < BUF_THRESHOLD && !this.fetching) {
            this.fetching = true;
            this.fetch();
        }

        const ret: IVector[] | undefined = this.buffer.shift();

        if (ret === undefined) {
            return null;
        } else {
            return ret;
        }
    }

}
