/**
 * @file Base interface definition for rendering pipeline elements.
 *
 * @license
 * Copyright (c) 2020 Felix Kopp <felix.kopp@stud.th-deg.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import { Object3D, Scene, Camera } from "three";

/**
 * Base interface for anything that is part of the rendering pipeline.
 */
export interface IRenderModule {

    isInitialized(): boolean;

    /** Do any initialization work necessary, if required. */
    onInit(): void;

    /**
     * Create all 3D objects controlled by this module.
     *
     * @returns The object or an array of all objects this hook is handling.
     */
    buildObjects(): (Object3D | undefined)[];

    /**
     * Callback for handling any changes to dynamic objects before drawing the
     * scene.
     *
     * @param time The timestamp of the current frame, in milliseconds.
     * @param delta The difference between the current and the previous render
     * timestamp in milliseconds.
     */
    onRender(time: number, delta: number): void;

    /**
     * Callback for changes to the rendering viewport size.
     *
     * @param w The new width, in pixels.
     * @param h The new height, in pixels.
     */
    onResize(w: number, h: number): void;

}
