/**
 * @file Abstract base class for all render modules.
 *
 * @license
 * Copyright (c) 2020 Felix Kopp <felix.kopp@stud.th-deg.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import { Object3D } from 'three';

import { IRenderModule } from '../interfaces/render-module';
import { IARRenderer } from '../interfaces/renderer';

/**
 * Abstact base class for all render modules.
 *
 * Render modules are simple callback-driven elements that can take care of one
 * or more 3D objects.  Before every frame, the
 * `{@link AbstractRenderModule#onRender}` callback is invoked to allow the module
 * to manipulate its objects if appropriate.
 */
export abstract class AbstractRenderModule implements IRenderModule {

    /** The renderer. */
    private renderer?: IARRenderer;

    private _isInitialized: boolean = false;

    /**
     * Create a new render module.
     *
     * @param renderer The renderer.
     * @param scene The scene.
     * @param camera The camera.
     */
    public constructor() { }

    public isInitialized(): boolean {
        return this._isInitialized;
    }

    /**
     * Return the main renderer.
     *
     * @returns The main renderer.
     */
    protected getRenderer(): IARRenderer {
        if (this.renderer === undefined) {
            throw new Error('Cannot call getRenderer() before onInit()');
        }

        return this.renderer;
    }

    /**
     * Internal initialization routine that calls `{@link #onInit}.
     * Do not override this.
     *
     * @param renderer The renderer.
     */
    public init(renderer: IARRenderer) {
        this.renderer = renderer;
        this.onInit();
        this._isInitialized = true;
    }

    /**
     * @override
     * @inheritdoc
     */
    public onInit(): void { }

    /**
     * @override
     * @inheritdoc
     */
    public abstract buildObjects(): (Object3D | undefined)[];

    /**
     * @override
     * @inheritdoc
     */
    public abstract onRender(time: number, delta: number): void;

    /**
     * @override
     * @inheritdoc
     */
    public onResize(w: number, h: number): void { }

}
