/**
 * @file Provides the main renderer.
 *
 * @license
 * Copyright (c) 2020 Felix Kopp <felix.kopp@stud.th-deg.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import {
    Color,
    WebGLRenderer,
    Scene,
    Camera,
    Object3D,
} from 'three';

import { IRenderModule } from '../interfaces/render-module';
import { IARRenderer } from '../interfaces/renderer';
import { AbstractRenderModule } from './abstract-render-module';

/**
 * The main renderer.
 */
export class ARRenderer implements IRenderModule, IARRenderer {

    /** An array of all rendering modules. */
    private readonly modules: AbstractRenderModule[] = [];
    /** The last time (in milliseconds) a frame was drawn. */
    private lastRenderTime: number = -1;

    private _isInitialized: boolean = false;

    /** The camera. */
    public readonly camera: Camera;
    /** The scene. */
    public readonly scene: Scene;
    /** The renderer. */
    public readonly glRenderer: WebGLRenderer;

    /** Create a new renderer. */
    constructor() {
        this.glRenderer = new WebGLRenderer({
            antialias: true,
            alpha: true,
        });
        this.camera = new Camera();
        this.scene = new Scene();
        this.onInit();
    }

    /**
     * @override
     * @inheritdoc
     */
    public addModule(mod: AbstractRenderModule) {
        if (mod.isInitialized()) {
            throw new Error(
                'RenderModule already initialized, do you have multiple instances of ARRenderer?'
            );
        }

        if (this.modules.indexOf(mod) === -1) {
            if (this.isInitialized()) {
                mod.init(this);
            }
            this.modules.push(mod);
        } else {
            throw new Error('Cannot add the same module multiple times');
        }
    }

    /**
     * @override
     * @inheritdoc
     */
    public getCamera(): Camera {
        return this.camera;
    }

    /**
     * @override
     * @inheritdoc
     */
    public getDomElement(): HTMLCanvasElement {
        return this.glRenderer.domElement;
    }

    /**
     * @override
     * @inheritdoc
     */
    public getScene(): Scene {
        return this.scene;
    }

    /**
     * @override
     * @inheritdoc
     */
    public isInitialized(): boolean {
        return this._isInitialized;
    }

    /**
     * Internal callback to `requestAnimationFrame()` aka the render loop.
     *
     * @param currentTime The timestamp (in milliseconds) for which to draw the
     * frame.
     */
    private drawAnimationFrame(currentTime: number) {
        requestAnimationFrame(newTime => this.drawAnimationFrame(newTime));

        const delta: number = Math.min(200, currentTime - this.lastRenderTime);
        this.lastRenderTime = currentTime;

        this.onRender(currentTime, delta);
    }

    /**
     * Start the rendering loop.
     */
    public start() {
        this.buildObjects();
        this.lastRenderTime = Date.now() - 1000 / 60;
        requestAnimationFrame(time => this.drawAnimationFrame(time));
    }

    /**
     * @override
     * @inheritdoc
     */
    public onInit() {
        this._isInitialized = true;
        this.scene.add(this.camera);

        this.glRenderer.setClearColor(new Color('lightgrey'), 0);
        this.glRenderer.setSize(window.innerWidth, window.innerHeight);
        this.glRenderer.setPixelRatio(1);
        this.glRenderer.domElement.classList.add('webar-renderer');
        document.body.appendChild(this.getDomElement());

        window.addEventListener(
            'resize',
            () => this.onResize(window.innerWidth, window.innerHeight)
        );
        for (const mod of this.modules) {
            if (!mod.isInitialized()) {
                mod.init(this);
            }
        }
    }

    /**
     * @override
     * @inheritdoc
     */
    public onRender(time: number, delta: number) {
        this.glRenderer.render(this.scene, this.camera);
        for (const mod of this.modules) {
            mod.onRender(time, delta);
        }
    }

    /**
     * @override
     * @inheritdoc
     */
    public onResize(w: number, h: number) {
        this.glRenderer.setSize(w, h);
        this.glRenderer.setPixelRatio(1);
        for (const mod of this.modules) {
            mod.onResize(w, h);
        }
    }

    /**
     * @override
     * @inheritdoc
     */
    public buildObjects(): Object3D[] {
        let objects: Object3D[] = [];

        for (const mod of this.modules) {
            const modObjects: (Object3D | undefined)[] = mod.buildObjects();

            for (const obj of modObjects) {
                if (obj !== undefined) {
                    objects.push(obj);
                }
            }
        }

        this.scene.add(...objects);
        return objects;
    }

}
