/**
 * @file Rendering module for the ar.js camera controller.
 *
 * @license
 * Copyright (c) 2020 Felix Kopp <felix.kopp@stud.th-deg.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import { ArToolkitSource, ArToolkitContext, ArMarkerControls } from '../../util/arjs-wrapper';
import { AbstractRenderModule } from '../abstract-render-module';

/**
 * Rendering module for ARToolkit.
 */
export class ARRenderModule extends AbstractRenderModule {

    /** The image source for ARToolkit. */
    private source: any;
    /** The rendering context for ARToolkit. */
    private context: any;
    /** The control element for adjusting the camera position. */
    private markerControls: any;

    /**
     * @override
     * @inheritdoc
     */
    public buildObjects(): undefined[] {
        /* We just need change the camera position */
        return [];
    }

    /**
     * @override
     * @inheritdoc
     */
    public onInit(): void {
        console.log('ar renderer init');
        this.source = new ArToolkitSource({
            sourceType: 'webcam',
        });
        this.context = new ArToolkitContext({
            cameraParametersUrl: '/static/camera_params.dat',
            detectionMode: 'mono',
        });

        this.getRenderer().getScene().visible = false;

        this.context.init(() => {
            this.getRenderer().getCamera().projectionMatrix.copy(
                this.context.getProjectionMatrix()
            );
        });
        this.source.init(
            () => this.onResize(window.innerWidth, window.innerHeight)
        );

        this.markerControls = new ArMarkerControls(
            this.context,
            this.getRenderer().getCamera(),
            {
                type: 'pattern',
                patternUrl: '/static/patt.hiro',
                changeMatrixMode: 'cameraTransformMatrix',
            }
        );
    }

    /**
     * @override
     * @inheritdoc
     */
    public onRender(time: number, delta: number): void {
        if (this.source.ready === true) {
            this.context.update(this.source.domElement);
            this.getRenderer().getScene().visible =
                this.getRenderer().getCamera().visible;
        }
    }

    /**
     * @override
     * @inheritdoc
     */
    public onResize(w: number, h: number): void {
        this.source.onResizeElement();
        this.source.copyElementSizeTo(this.getRenderer().getDomElement());

        if (this.context.arController !== null) {
            this.source.copyElementSizeTo(this.context.arController.canvas);
        }
    }

}
