/**
 * @file Provides a demo module that draws a cube onto the AR marker.
 *
 * @license
 * Copyright (c) 2020 Felix Kopp <felix.kopp@stud.th-deg.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import {
    BoxGeometry,
    Object3D,
    MeshNormalMaterial,
    DoubleSide,
    Mesh,
    TorusKnotGeometry
} from 'three';

import { AbstractRenderModule } from '../abstract-render-module';

/**
 * Dummy class that renders a cube containing a torus sphere on the AR marker.
 */
export class DemoCubeRenderModule extends AbstractRenderModule {

    private boxMesh?: Object3D;
    private torusMesh?: Object3D;

    /**
     * @override
     * @inheritdoc
     */
    public onInit() {
        console.log('demo cube init');
        const boxGeom = new BoxGeometry(1, 1, 1);
        const boxMaterial = new MeshNormalMaterial({
            transparent: true,
            opacity: 0.5,
            side: DoubleSide,
        });
        this.boxMesh = new Mesh(boxGeom, boxMaterial);
        this.boxMesh.position.y = boxGeom.parameters.height / 2;

        const torusGeom = new TorusKnotGeometry(0.3, 0.1, 64, 46);
        const torusMaterial = new MeshNormalMaterial();
        this.torusMesh = new Mesh(torusGeom, torusMaterial);
        this.torusMesh.position.y = 0.5;
    }

    /**
     * @override
     * @inheritdoc
     */
    public buildObjects(): (Object3D | undefined)[] {
        return [
            this.boxMesh,
            this.torusMesh,
        ];
    }

    /**
     * @override
     * @inheritdoc
     */
    public onRender(time: number, delta: number) {
        if (this.torusMesh === undefined) {
            return;
        }

        this.torusMesh.rotation.x += Math.PI * (delta / 1000);
    }

}
