/**
 * @file The actual vector field rendering module.
 * @license
 * Copyright (c) 2020 Felix Kopp <felix.kopp@stud.th-deg.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

import { Object3D, Line, LineBasicMaterial, Vector3, Geometry, Group } from 'three';

import { AbstractRenderModule } from '../abstract-render-module';
import { IVectorDataSource } from '../../interfaces/vector-data-source';
import { VectorDataSource } from '../../data/vector-data-source';
import { IVector, IDrawableVector } from '../../interfaces/vect';

const VERT_INDEX_POS: number = 0;
const VERT_INDEX_DIR: number = 1;

export interface VectorFieldRenderModuleOpts {
    /** Data retrieval options. */
    fetch?: {
        /** How many frames to skip before fetching new data. */
        interval?: number,
        /** The data uri where we can get our vectors. */
        uri?: string,
    },
    /** A scaling factor for all arrows. */
    scale?: number,
}

export class VectorFieldRenderModule extends AbstractRenderModule {

    /** The data source to get vectors from. */
    private readonly dataSource: IVectorDataSource;
    /* The amount of frames since the last data refresh. */
    private lastFetch: number = 0;
    private fetchInterval: number;

    private vectors: IDrawableVector[] = [];
    private objGroup: Group;

    public lineScale: number = 0.33;

    /**
     * Create a new vector field.
     *
     * @param opts Options.
     */
    public constructor(opts?: VectorFieldRenderModuleOpts) {
        super();

        /* TODO: Parse config sanely */
        if (opts) {
            if (opts.fetch) {
                this.fetchInterval = opts.fetch.interval ? opts.fetch.interval : 60;
                this.dataSource = new VectorDataSource(
                    opts.fetch.uri ? opts.fetch.uri : '/arrows'
                );
            } else {
                this.fetchInterval = 60;
                this.dataSource = new VectorDataSource('/arrows');
            }
            if (opts.scale) {
                this.lineScale = opts.scale;
            }
        } else {
            this.fetchInterval = 60;
            this.dataSource = new VectorDataSource('/arrows');
        }

        this.objGroup = new Group();
        this.objGroup.position.y = 1;
        console.log(this.lineScale);
    }

    /**
     * @override
     * @inheritdoc
     */
    public buildObjects(): Object3D[] {
        return [ this.objGroup ];
    }

    /**
     * @override
     * @inheritdoc
     */
    public onRender(time: number, delta: number) {
        if (++this.lastFetch === this.fetchInterval) {
            this.lastFetch = 0;
            const newPositions: IVector[] | null = this.dataSource.getNext();
            if (newPositions !== null) {
                this.updatePositions(newPositions);
            }
        }
    }

    /**
     * Update the line position and colors by reinstantiating all graphics
     * objects.
     *
     * @param positions The new positions.
     */
    private updatePositions(positions: IVector[]) {
        let vect: IVector | undefined;
        let i: number = 0;

        while (vect = positions.pop()) {
            if (i < this.vectors.length) {
                this.objGroup.remove(this.vectors[i].line);
            }
            this.vectors[i] = this.createDrawableVect(vect);

            i++;
        }

        while (i < this.vectors.length) {
            this.objGroup.remove(this.vectors[i].line);
            i++;
        }
    }

    /**
     * Generate a geometry, material and 3d object for the specified vector.
     *
     * @param vect The vector to build.
     */
    private createDrawableVect(vect: IVector): IDrawableVector {
        const geom = new Geometry();
        geom.vertices[VERT_INDEX_POS] = new Vector3(
            vect.base.x * this.lineScale,
            vect.base.y * this.lineScale,
            vect.base.z * this.lineScale
        );
        geom.vertices[VERT_INDEX_DIR] = new Vector3(
            vect.base.x * this.lineScale + vect.direction.x * this.lineScale,
            vect.base.y * this.lineScale + vect.direction.y * this.lineScale,
            vect.base.z * this.lineScale + vect.direction.z * this.lineScale
        );
        const line = new Line(
            geom,
            this.getMaterialFromLen(geom.vertices[VERT_INDEX_DIR].length()),
        );

        this.objGroup.add(line);

        return {
            geom: geom,
            line: line,
        };
    }

    /**
     * Generate a material of appropriate color for a vector.
     * Longer vectors are more red, shorter ones more green.
     *
     * @param len The vector length.
     */
    private getMaterialFromLen(len: number): LineBasicMaterial {
        // min: 120° (len 0)
        // max: 0° (len 10)

        let hue: number = 120 - (len / (2 * this.lineScale)) * 120
        if (hue < 0) {
            hue = 0;
        } else if (hue > 120) {
            hue = 120;
        }

        /* TODO: Compute actual color */
        return new LineBasicMaterial({
            color: `hsl(${Math.round(hue)}, 100%, 50%)`,
            linewidth: 2,
        });
    }

}
